from __future__ import annotations

from pathlib import Path

import pytest

from ..dron import load_jobs


def test_load_jobs_basic(tmp_path: Path) -> None:
    tpath = Path(tmp_path) / 'drontab.py'
    tpath.write_text(
        '''
from typing import Iterator

from dron.api import job, Job


def jobs() -> Iterator[Job]:
    job3 = job(
        '03:10',
        ['/path/to/command.py', 'some', 'args', '3'],
        unit_name='job3',
    )
    job1 = job(
        '01:10',
        ['/path/to/command.py', 'some', 'args', '1'],
        unit_name='job1',
    )
    yield job1
    yield job(
        '02:10',
        ['/path/to/command.py', 'some', 'args', '2'],
        unit_name='job2',
    )
    yield job3

'''
    )
    loaded = list(load_jobs(tabfile=tpath, ppath=tmp_path))
    [job1, job2, job3] = loaded

    assert job1.when == '01:10'
    assert job1.command == ['/path/to/command.py', 'some', 'args', '1']
    assert job1.unit_name == 'job1'

    assert job2.when == '02:10'
    assert job2.command == ['/path/to/command.py', 'some', 'args', '2']
    assert job2.unit_name == 'job2'

    assert job3.when == '03:10'
    assert job3.command == ['/path/to/command.py', 'some', 'args', '3']
    assert job3.unit_name == 'job3'


def test_load_jobs_dupes(tmp_path: Path) -> None:
    tpath = Path(tmp_path) / 'drontab.py'
    tpath.write_text(
        '''
from typing import Iterator

from dron.api import job, Job

def jobs() -> Iterator[Job]:
    yield job('00:00', 'echo', unit_name='job3')
    yield job('00:00', 'echo', unit_name='job1')
    # whoops! duplicate job name
    yield job('00:00', 'echo', unit_name='job3')
'''
    )
    with pytest.raises(AssertionError):
        _loaded = list(load_jobs(tabfile=tpath, ppath=tmp_path))


def test_jobs_auto_naming(tmp_path: Path) -> None:
    tpath = Path(tmp_path) / 'drontab.py'
    tpath.write_text(
        '''
from typing import Iterator

from dron.api import job, Job


job2 = job(
    '00:02',
    'echo',
)


def job_maker(when) -> Job:
    return job(when, 'echo job maker', stacklevel=2)


def jobs() -> Iterator[Job]:
    job_1 = job('00:01',
        'echo',
    )
    yield job2
    yield job('00:00', 'echo', unit_name='job_named')
    yield job_1
    job4 = \
       job('00:04', 'echo')
    job5     = job_maker('00:05')
    yield job5
    yield job4
'''
    )
    loaded = list(load_jobs(tabfile=tpath, ppath=tmp_path))
    (job2, job_named, job_1, job5, job4) = loaded
    assert job_1.unit_name == 'job_1'
    assert job_1.when == '00:01'
    assert job2.unit_name == 'job2'
    assert job2.when == '00:02'
    assert job_named.unit_name == 'job_named'
    assert job_named.when == '00:00'
    assert job4.unit_name == 'job4'
    assert job4.when == '00:04'
    assert job5.unit_name == 'job5'
    assert job5.when == '00:05'



from __future__ import annotations

import argparse
import os
import shutil
import subprocess
import sys
from pathlib import Path
from pprint import pprint
from tempfile import TemporaryDirectory

import click

from . import launchd, systemd
from .api import UnitName
from .common import (
    IS_SYSTEMD,
    MANAGED_MARKER,
    MonitorParams,
    Unit,
    escape,
    logger,
    print_monitor,
)
from .dron import (
    DRONTAB,
    apply,
    do_lint,
    drontab_dir,
    get_entries_for_monitor,
    load_jobs,
    manage,
    managed_units,
)


def cmd_edit() -> None:
    drontab = DRONTAB
    if not drontab.exists():
        if click.confirm(f"tabfile {drontab} doesn't exist. Create?", default=True):
            drontab.write_text(
                '''\
#!/usr/bin/env python3
from dron.api import job

def jobs():
    # yield job(
    #     'hourly',
    #     '/bin/echo 123',
    #     unit_name='test_unit'
    # )
    pass
'''.lstrip()
            )
        else:
            raise RuntimeError()

    editor = os.environ.get('EDITOR')
    if editor is None:
        logger.warning('No EDITOR! Fallback to nano')
        editor = 'nano'

    with TemporaryDirectory() as tdir:
        tpath = Path(tdir) / 'drontab'
        shutil.copy2(drontab, tpath)

        orig_mtime = tpath.stat().st_mtime
        while True:
            res = subprocess.run([editor, str(tpath)], check=True)

            new_mtime = tpath.stat().st_mtime
            if new_mtime == orig_mtime:
                logger.warning('No notification made')
                return

            ex: Exception | None = None
            try:
                state = do_lint(tabfile=tpath)
            except Exception as e:
                logger.exception(e)
                ex = e
            else:
                try:
                    manage(state=state)
                except Exception as ee:
                    logger.exception(ee)
                    ex = ee
            if ex is not None:
                if click.confirm('Got errors. Try again?', default=True):
                    continue
                raise ex

            drontab.write_text(tpath.read_text())  # handles symlinks correctly
            logger.info(f"Wrote changes to {drontab}. Don't forget to commit!")
            break

        # TODO show git diff?
        # TODO perhaps allow to carry on regardless? not sure..
        # not sure how much we can do without modifying anything...


def cmd_lint(tabfile: Path) -> None:
    _state = do_lint(tabfile)
    logger.info('all good')


def cmd_apply(tabfile: Path) -> None:
    apply(tabfile)


def cmd_print(*, tabfile: Path, pretty: bool) -> None:
    dtab_dir = Path(drontab_dir())
    jobs = list(load_jobs(tabfile=tabfile, ppath=dtab_dir))

    if pretty:
        import tabulate

        items = [
            {
                'UNIT': job.unit_name,
                'SCHEDULE': job.when,
                'COMMAND': escape(job.command),
            }
            for job in jobs
        ]
        print(tabulate.tabulate(items, headers="keys"))
    else:
        for j in jobs:
            print(j)


def cmd_run(*, unit: Unit, do_exec: bool) -> None:
    if IS_SYSTEMD:
        return systemd.cmd_run(unit=unit, do_exec=do_exec)
    else:
        return launchd.cmd_run(unit=unit, do_exec=do_exec)


def cmd_past(unit: Unit) -> None:
    if IS_SYSTEMD:
        return systemd.cmd_past(unit)
    else:
        return launchd.cmd_past(unit)


# TODO think if it's worth integrating with timers?
def cmd_monitor(params: MonitorParams) -> None:
    managed = list(managed_units(with_body=False))  # body slows down this call quite a bit
    if len(managed) == 0:
        logger.warning('no managed units!')

    logger.debug('starting monitor...')

    entries = get_entries_for_monitor(managed=managed, params=params)
    print_monitor(entries)


# TODO test it on CI?
def _drontab_example() -> str:
    return '''
from dron.api import job

# at the moment you're expected to define jobs() function that yields jobs
# in the future I might add more mechanisms
def jobs():
    # simple job that doesn't do much
    yield job(
        'daily',
        '/home/user/scripts/run-borg /home/user',
        unit_name='borg-backup-home',
    )

    yield job(
        'daily',
        'linkchecker https://beepb00p.xyz',
        unit_name='linkchecker-beepb00p',
    )

    # drontab is simply python code!
    # so if you're annoyed by having to rememver Systemd syntax, you can use a helper function
    def every(*, mins: int) -> str:
        return f'*:0/{mins}'

    # make sure my website is alive, it will send local email on failure
    yield job(
        every(mins=10),
        'ping https://beepb00p.xyz',
        unit_name='ping-beepb00p',
    )
'''.lstrip()


def make_parser() -> argparse.ArgumentParser:
    from .common import VerifyOff

    def add_verify(p: argparse.ArgumentParser) -> None:
        # specify in readme???
        # would be nice to use external checker..
        # https://github.com/systemd/systemd/issues/8072
        # https://unix.stackexchange.com/questions/493187/systemd-under-ubuntu-18-04-1-fails-with-failed-to-create-user-slice-serv
        p.add_argument('--no-verify', action=VerifyOff, nargs=0, help='Skip systemctl verify step')

    p = argparse.ArgumentParser(
        prog='dron',
        description='''
dron -- simple frontend for Systemd, inspired by cron.

- *d* stands for 'Systemd'
- *ron* stands for 'cron'

dron is my attempt to overcome things that make working with Systemd tedious
'''.lstrip(),
        formatter_class=lambda prog: argparse.RawTextHelpFormatter(prog, width=100),
    )
    # TODO ugh. when you type e.g. 'dron apply', help format is wrong..
    example = ''.join(': ' + l for l in _drontab_example().splitlines(keepends=True))
    # TODO begin_src python maybe?
    p.epilog = f'''
* What does it do?
In short, you type ~dron edit~ and edit your config file, similarly to ~crontab -e~:

{example}

After you save your changes and exit the editor, your drontab is checked for syntax and applied

- if checks have passed, your jobs are mapped onto Systemd units and started up
- if there are potential errors, you are prompted to fix them before retrying

* Why?
In short, because I want to benefit from the heavy lifting that Systemd does: timeouts, resource management, restart policies, powerful scheduling specs and logging,
while not having to manually manipulate numerous unit files and restart the daemon all over.

I elaborate on what led me to implement it and motivation [[https://beepb00p.xyz/scheduler.html#what_do_i_want][here]]. Also:

- why not just use [[https://beepb00p.xyz/scheduler.html#cron][cron]]?
- why not just use [[https://beepb00p.xyz/scheduler.html#systemd][systemd]]?
    '''

    p.add_argument(
        '--marker', required=False, help=f'Use custom marker instead of default `{MANAGED_MARKER}`. Useful for developing/testing.'
    )

    def add_tabfile_arg(p: argparse.ArgumentParser) -> None:
        p.add_argument('tabfile', type=Path, nargs='?')

    sp = p.add_subparsers(dest='mode')

    ### actions on drontab file
    edit_parser = sp.add_parser('edit', help="Edit drontab (like 'crontab -e')")
    add_verify(edit_parser)

    apply_parser = sp.add_parser('apply', help="Apply drontab (like 'crontab' with no args)")
    add_tabfile_arg(apply_parser)
    add_verify(apply_parser)
    # TODO --force?
    lint_parser = sp.add_parser('lint', help="Check drontab (no 'crontab' alternative, sadly!)")
    add_tabfile_arg(lint_parser)
    add_verify(lint_parser)

    print_parser = sp.add_parser('print', help="Parse and print drontab")
    add_tabfile_arg(print_parser)
    print_parser.add_argument('--pretty', action='store_true')
    ###

    ### actions on managed jobs
    debug_parser = sp.add_parser('debug', help='Print some debug info')

    uninstall_parser = sp.add_parser('uninstall', help="Uninstall all managed jobs")
    add_verify(uninstall_parser)

    run_parser = sp.add_parser('run', help='Run the job right now, ignoring the timer')
    run_parser.add_argument('unit', type=str, nargs='?')  # TODO add shell completion?
    run_parser.add_argument('--exec', action='store_true', dest='do_exec', help='Run directly, not via systemd/launchd')

    past_parser = sp.add_parser('past', help='List past job runs')
    past_parser.add_argument('unit', type=str, nargs='?')  # TODO add shell completion?
    ###

    ### misc actions
    mp = sp.add_parser('monitor', help='Monitor services/timers managed by dron')
    mp.add_argument('-n', type=int, default=1, help='refresh every n seconds')
    mp.add_argument('--once', action='store_true', help='only call once')
    mp.add_argument('--rate', action='store_true', help='Display success rate (unstable and potentially slow)')
    mp.add_argument('--command', action='store_true', help='Display command')
    ###
    return p


def main() -> None:
    from .cli import make_parser

    p = make_parser()
    args = p.parse_args()

    marker: str | None = args.marker
    if marker is not None:
        from . import common

        common.MANAGED_MARKER = marker

    mode: str = args.mode

    def tabfile_or_default() -> Path:
        tabfile = args.tabfile
        if tabfile is None:
            tabfile = DRONTAB
        return tabfile

    def prompt_for_unit() -> UnitName:
        from prompt_toolkit import PromptSession
        from prompt_toolkit.completion import WordCompleter

        # TODO print options
        managed = list(managed_units(with_body=False))
        units = [x.unit_file.stem for x in managed]

        print('Units under dron:', file=sys.stderr)
        for u in units:
            print(f'- {u}', file=sys.stderr)

        completer = WordCompleter(units, ignore_case=True)
        session = PromptSession("Select a unit: ", completer=completer)  # type: ignore[var-annotated]
        selected = session.prompt()
        return selected

    if mode == 'edit':
        cmd_edit()
    elif mode == 'apply':
        tabfile = tabfile_or_default()
        cmd_apply(tabfile)
    elif mode == 'lint':
        tabfile = tabfile_or_default()
        cmd_lint(tabfile)
    elif mode == 'print':
        tabfile = tabfile_or_default()

        from .cli import cmd_print  # lazy due to circular import

        cmd_print(tabfile=tabfile, pretty=args.pretty)
    elif mode == 'debug':
        managed = managed_units(with_body=False)  # TODO not sure about body
        for x in managed:
            pprint(x, stream=sys.stderr)
    elif mode == 'uninstall':
        click.confirm('Going to remove all dron managed jobs. Continue?', default=True, abort=True)
        with TemporaryDirectory() as td:
            empty = Path(td) / 'empty'
            empty.write_text(
                '''\
def jobs():
    yield from []
'''
            )
            cmd_apply(empty)
    elif mode == 'run':
        unit = args.unit if args.unit is not None else prompt_for_unit()
        do_exec = args.do_exec
        cmd_run(unit=unit, do_exec=do_exec)
    elif mode == 'past':
        unit = args.unit if args.unit is not None else prompt_for_unit()
        cmd_past(unit=unit)
    elif mode == 'monitor':
        once = args.once

        params = MonitorParams(
            with_success_rate=args.rate,
            with_command=args.command,
        )

        if once:
            # fallback on old style monitor for now?
            # this can be quite useful for grepping etc..
            cmd_monitor(params=params)
        else:
            from .monitor import MonitorApp

            app = MonitorApp(monitor_params=params, refresh_every=args.n)
            app.run()
    else:
        logger.error(f'Unknown mode: {mode}')
        p.print_usage(sys.stderr)
        sys.exit(1)



#!/usr/bin/env python3
import argparse
import shlex
import sys
from pathlib import Path
from subprocess import PIPE, STDOUT, Popen
from typing import Iterator, NoReturn

from loguru import logger

LOG_DIR = Path('~/Library/Logs/dron').expanduser()


def main() -> NoReturn:
    p = argparse.ArgumentParser()
    p.add_argument('--notify', action='append')
    p.add_argument('--job', required=True)
    # hmm, this doesn't work with keyword args??
    # p.add_argument('cmd', nargs=argparse.REMAINDER)
    args, rest = p.parse_known_args()

    assert rest[0] == '--', rest
    cmd = rest[1:]

    notify_cmds = [] if args.notify is None else args.notify
    job = args.job

    LOG_DIR.mkdir(parents=True, exist_ok=True)
    log_file = LOG_DIR / f'{job}.log'

    logger.add(log_file, rotation='100 MB')  # todo configurable? or rely on osx rotation?

    # hmm, a bit crap transforming everything to stdout? but not much we can do?
    captured_log = []
    try:
        with Popen(cmd, stdout=PIPE, stderr=STDOUT) as po:
            out = po.stdout
            assert out is not None
            for line in out:
                captured_log.append(line)
                sys.stdout.buffer.write(line)
        rc = po.poll()

        if rc == 0:
            # short circuit
            sys.exit(0)
    except Exception as e:
        # Popen istelf still fail due to permission denied or something
        logger.exception(e)
        captured_log.append(str(e).encode('utf8'))
        rc = 123


    def payload() -> Iterator[bytes]:
        yield f"exit code: {rc}\n".encode()
        yield b'command: \n'
        yield (' '.join(map(shlex.quote, cmd)) + '\n').encode('utf8')
        yield f'log file: {log_file}\n'.encode()
        yield b'\n'
        yield b'output (stdout + stderr):\n\n'
        # TODO shit -- if multiple notifications, can't use generator for captured_log
        # unless we notify simultaneously?
        yield from captured_log

    for line in payload():
        logger.info(line.decode('utf8').rstrip('\n'))  # meh

    for notify_cmd in notify_cmds:
        logger.info(f'notifying: {notify_cmd}')
        try:
            with Popen(notify_cmd, shell=True, stdin=PIPE, stdout=PIPE, stderr=PIPE) as po:
                sin = po.stdin
                assert sin is not None
                for line in payload():
                    sin.write(line)
                (sout, serr) = po.communicate()
                for l in sout.decode('utf8').splitlines():
                    logger.debug(l)
                for l in serr.decode('utf8').splitlines():
                    logger.debug(l)
            assert po.poll() == 0, notify_cmd
        except Exception as e:
            logger.error(f'notificaiton failed: {notify_cmd}')
            logger.exception(e)

    sys.exit(rc)


if __name__ == '__main__':
    main()



from __future__ import annotations

import json
import os
import re
import shlex
import shutil
from datetime import datetime, timedelta, timezone
from functools import lru_cache
from itertools import groupby
from pathlib import Path
from subprocess import PIPE, Popen, run
from tempfile import TemporaryDirectory
from typing import Any, Iterator, Sequence

from zoneinfo import ZoneInfo

from .api import (
    OnFailureAction,
    When,
)
from .common import (
    MANAGED_MARKER,
    Body,
    Command,
    MonitorEntry,
    MonitorParams,
    State,
    TimerSpec,
    Unit,
    UnitState,
    datetime_aware,
    escape,
    is_managed,
    logger,
)


def _is_missing_systemd() -> str | None:
    has_systemd = shutil.which('systemctl') is not None
    if not has_systemd:
        return "systemd not available, running under docker or osx"
    return None


def _systemctl(*args: Path | str) -> list[Path | str]:
    return ['systemctl', '--user', *args]


def managed_header() -> str:
    return f'''
# {MANAGED_MARKER}
# If you do any manual changes, they will be overridden on the next dron run
'''.lstrip()


# TODO how to come up with good implicit job name?
# TODO do we need a special target for dron?
def timer(*, unit_name: str, when: When) -> str:
    spec: TimerSpec
    if isinstance(when, str):
        spec = {'OnCalendar': when}
    else:
        spec = when

    specs = '\n'.join(f'{k}={v}' for k, v in spec.items())

    return f'''
{managed_header()}
[Unit]
Description=Timer for {unit_name} {MANAGED_MARKER}

[Timer]
{specs}

[Install]
WantedBy=timers.target
'''.lstrip()


# TODO add Restart=always and RestartSec?
# TODO allow to pass extra args
def service(
        *,
        unit_name: str,
        command: Command,
        on_failure: Sequence[OnFailureAction],
        **kwargs: str,
) -> str:
    # TODO not sure if something else needs to be escaped for ExecStart??
    # todo systemd-escape? but only can be used for names

    # ok OnFailure is quite annoying since it can't take arguments etc... seems much easier to use ExecStopPost
    # (+ can possibly run on success too that way?)
    # https://unix.stackexchange.com/a/441662/180307
    cmd = escape(command)

    exec_stop_posts = [
        f"ExecStopPost=/bin/sh -c 'if [ $$EXIT_STATUS != 0 ]; then {action}; fi'"
        for action in on_failure
    ]

    sections: dict[str, list[str]] = {}
    sections['[Unit]'] = [f'''
Description=Service for {unit_name} {MANAGED_MARKER}
'''.strip()]

    sections['[Service]'] = [
        f'ExecStart={cmd}',
        *exec_stop_posts,
    ]

    for k, value in kwargs.items():
        # ideally it would have section name
        m = re.search(r'(\[\w+\])(.*)', k)
        if m is not None:
            section = m.group(1)
            key = m.group(2)
        else:
            # 'legacy' behaviour, by default put into [Service]
            section = '[Service]'
            key = k
        if section not in sections:
            sections[section] = []
        sections[section].append(f'{key}={value}')

    res = managed_header()
    for section_name, lines in sections.items():
        res += '\n\n' + '\n'.join([section_name, *lines])
    res += '\n'

    return res


def test_managed() -> None:
    skip_if_no_systemd()
    from .dron import verify_unit

    assert is_managed(timer(unit_name='whatever', when='daily'))

    custom = '''
[Service]
ExecStart=/bin/echo 123
'''
    verify_unit(unit_name='other.service', body=custom)  # precondition
    assert not is_managed(custom)


def verify_units(pre_units: list[tuple[Unit, Body]]) -> None:
    # ugh. systemd-analyze takes about 0.2 seconds for each unit for some reason
    # oddly enough, in bulk it works just as fast :thinking_face:
    # also doesn't work in parallel (i.e. parallel processes)
    # that ends up with some weird errors trying to connect to socket
    with TemporaryDirectory() as _tdir:
        tdir = Path(_tdir)
        for unit, body in pre_units:
            (tdir / unit).write_text(body)
        res = run(['systemd-analyze', '--user', 'verify', *tdir.glob('*')], capture_output=True, check=False)
        # ugh. apparently even exit code 0 doesn't guarantee correct output??
        out = res.stdout.decode('utf8')
        err = res.stderr.decode('utf8')
        assert out == '', out
        if err == '':
            return

        err_lines = err.splitlines(keepends=True)
        unique_err_lines = []
        # uhh.. in bulk mode it spams with tons of 'Cannot add dependency job' for some reason
        # I guess it kinda treats everything as dependent on each other??
        # https://github.com/systemd/systemd/blob/b692ad36b99909453cf4f975a346e41d6afc68a0/src/core/transaction.c#L978
        for l in err_lines:
            if l not in unique_err_lines:
                unique_err_lines.append(l)
        err_lines = unique_err_lines

        if len(err_lines) == 0:
            return

        msg = f'failed checking , exit code {res.returncode}'
        logger.error(msg)
        logger.error('systemd-analyze output:')
        for line in err_lines:
            logger.error(line.strip())
        raise RuntimeError(msg)


def test_verify_systemd() -> None:
    skip_if_no_systemd()
    from .dron import verify_unit

    def fails(body: str) -> None:
        import pytest
        with pytest.raises(Exception):
            verify_unit(unit_name='whatever.service', body=body)

    def ok(body: str) -> None:
        verify_unit(unit_name='ok.service', body=body)

    ok(body='''
[Service]
ExecStart=/bin/echo 123
''')

    from .api import notify
    on_failure = (
        notify.email('test@gmail.com'),
        notify.desktop_notification,
    )
    ok(body=service(unit_name='alala', command='/bin/echo 123', on_failure=on_failure))

    # garbage
    fails(body='fewfewf')

    # no execstart
    fails(body='''
[Service]
StandardOutput=journal
''')

    fails(body='''
[Service]
ExecStart=yes
StandardOutput=baaad
''')


def _sd(s: str) -> str:
    return f'org.freedesktop.systemd1{s}'


class BusManager:
    def __init__(self) -> None:
        # unused-ignore because on macos there is no dbus (but this code is still running mypy on CI)
        from dbus import (  # type: ignore[import-untyped,import-not-found,unused-ignore]
            Interface,
            SessionBus,
        )
        self.Interface = Interface  # meh

        self.bus = SessionBus()  # note: SystemBus is for system-wide services
        systemd = self.bus.get_object(_sd(''), '/org/freedesktop/systemd1')
        self.manager = Interface(systemd, dbus_interface=_sd('.Manager'))

    def properties(self, u: Unit):
        service_unit = self.manager.GetUnit(u)
        service_proxy = self.bus.get_object(_sd(''), str(service_unit))
        return self.Interface(service_proxy, dbus_interface='org.freedesktop.DBus.Properties')

    @staticmethod  # meh
    def prop(obj, schema: str, name: str):
        return obj.Get(_sd(schema), name)

    @classmethod
    def exec_start(cls, props) -> Sequence[str]:
        dbus_exec_start = cls.prop(props, '.Service', 'ExecStart')
        return [str(x) for x in dbus_exec_start[0][1]]


def systemd_state(*, with_body: bool) -> State:
    bus = BusManager()
    states = bus.manager.ListUnits()  # ok nice, it's basically instant

    for state in states:
        name  = state[0]
        descr = state[1]
        if not is_managed(descr):
            continue

        # todo annoying, this call still takes some time... but whatever ok
        props = bus.properties(name)

        # useful for debugging, can also use .Service if it's not a timer
        # all_properties = props.GetAll(_sd('.Unit'))

        # stale = int(bus.prop(props, '.Unit', 'NeedDaemonReload')) == 1
        unit_file = Path(str(bus.prop(props, '.Unit', 'FragmentPath'))).resolve()
        body = unit_file.read_text() if with_body else None
        cmdline: Sequence[str] | None
        if '.timer' in name: # meh
            cmdline = None
        else:
            cmdline = BusManager.exec_start(props)

        yield UnitState(unit_file=unit_file, body=body, cmdline=cmdline)


def test_managed_units() -> None:
    skip_if_no_systemd()
    # TODO wonder if i'd be able to use launchd on ci...
    from .cli import cmd_monitor
    from .dron import managed_units

    # shouldn't fail at least
    list(managed_units(with_body=True))

    # TODO ugh. doesn't work on circleci, fails with
    # dbus.exceptions.DBusException: org.freedesktop.DBus.Error.BadAddress: Address does not contain a colon
    # todo maybe don't need it anymore with 20.04 circleci?
    if 'CI' not in os.environ:
        cmd_monitor(MonitorParams(with_success_rate=True, with_command=True))


def skip_if_no_systemd() -> None:
    import pytest
    reason = _is_missing_systemd()
    if reason is not None:
        pytest.skip(f'No systemd: {reason}')


_UTCMAX = datetime.max.replace(tzinfo=timezone.utc)


class MonitorHelper:
    def from_usec(self, usec) -> datetime_aware:
        u = int(usec)
        if u == 2 ** 64 - 1: # apparently systemd uses max uint64
            # happens if the job is running ATM?
            return _UTCMAX
        else:
            return datetime.fromtimestamp(u / 10 ** 6, tz=timezone.utc)

    @property
    @lru_cache  # noqa: B019
    def local_tz(self) -> ZoneInfo:
        try:
            # it's a required dependency, but still might fail in some weird environments?
            #   e.g. if zoneinfo information isn't available
            from tzlocal import get_localzone
            return get_localzone()
        except Exception as e:
            logger.error("Couldn't determine local timezone! Falling back to UTC")
            return ZoneInfo('UTC')


def get_entries_for_monitor(managed: State, *, params: MonitorParams) -> list[MonitorEntry]:
    # TODO reorder timers and services so timers go before?
    sd = lambda s: f'org.freedesktop.systemd1{s}'

    mon = MonitorHelper()

    UTCNOW = datetime.now(tz=timezone.utc)

    bus = BusManager()

    entries: list[MonitorEntry] = []
    names = sorted(s.unit_file.name for s in managed)
    uname = lambda full: full.split('.')[0]
    for k, _gr in groupby(names, key=uname):
        gr = list(_gr)
        # if timer is None, guess that means the job is always running?
        timer: str | None
        service: str
        if len(gr) == 2:
            [service, timer] = gr
        else:
            assert len(gr) == 1, gr
            [service] = gr
            timer = None

        if timer is not None:
            props = bus.properties(timer)
            cal   = bus.prop(props, '.Timer', 'TimersCalendar')
            next_ = bus.prop(props, '.Timer', 'NextElapseUSecRealtime')

            unit_props = bus.properties(service)
            # note: there is also bus.prop(props, '.Timer', 'LastTriggerUSec'), but makes more sense to use unit to account for manual runs
            last  = bus.prop(unit_props, '.Unit', 'ActiveExitTimestamp')

            schedule = cal[0][1]  # TODO is there a more reliable way to retrieve it??
            # todo not sure if last is really that useful..

            last_dt = mon.from_usec(last)
            next_dt = mon.from_usec(next_)
            nexts = next_dt.astimezone(mon.local_tz).replace(tzinfo=None, microsecond=0).isoformat()

            if next_dt == datetime.max:
                left_delta = timedelta(0)
            else:
                left_delta   = next_dt - UTCNOW
        else:
            left_delta = timedelta(0) # TODO
            last_dt = UTCNOW
            nexts = 'n/a'
            schedule = 'always'

        # TODO maybe format seconds prettier. dunno
        def fmt_delta(d: timedelta) -> str:
            # format to reduce constant countdown...
            ad = abs(d)
            # get rid of microseconds
            ad = ad - timedelta(microseconds=ad.microseconds)

            day    = timedelta(days=1)
            hour   = timedelta(hours=1)
            minute = timedelta(minutes=1)
            gt = False
            if ad > day:
                full_days  = ad // day
                hours = (ad % day) // hour
                ads = f'{full_days}d {hours}h'
                gt = True
            elif ad > minute:
                full_mins  = ad // minute
                ad = timedelta(minutes=full_mins)
                ads = str(ad)
                gt = True
            else:
                # show exact
                ads = str(ad)
            if len(ads) == 7:
                ads = '0' + ads # meh. fix missing leading zero in hours..
            ads = ('>' if gt else '') + ads
            return ads


        left   = f'{fmt_delta(left_delta)!s:<9}'
        if last_dt.timestamp() == 0:
            ago = 'never' # TODO yellow?
        else:
            passed_delta = UTCNOW - last_dt
            ago = str(fmt_delta(passed_delta))
        # TODO instead of hacking microsecond, use 'NOW' or something?

        props = bus.properties(service)
        # TODO some summary too? e.g. how often in failed
        # TODO make defensive?
        result = bus.prop(props, '.Service', 'Result')
        exec_start = BusManager.exec_start(props)
        assert exec_start is not None, service  # not None for services
        command = ' '.join(map(shlex.quote, exec_start)) if params.with_command else None
        _pid: int | None = int(bus.prop(props, '.Service', 'MainPID'))
        pid  = None if _pid == 0 else str(_pid)

        if params.with_success_rate:
            rate = _unit_success_rate(service)
            rates = f' {rate:.2f}'
        else:
            rates = ''

        status_ok = result == 'success'
        status = f'{result:<9} {ago:<8}{rates}'

        entries.append(MonitorEntry(
            unit=k,
            status=status,
            left=left,
            next=nexts,
            schedule=schedule,
            command=command,
            pid=pid,
            status_ok=status_ok,
        ))
    return entries


Json = dict[str, Any]
def _unit_logs(unit: Unit) -> Iterator[Json]:
    # TODO so do I need to parse logs to get failure stats? perhaps json would be more reliable
    cmd = f'journalctl --user -u {unit} -o json -t systemd --output-fields UNIT_RESULT,JOB_TYPE,MESSAGE'
    with Popen(cmd.split(), stdout=PIPE) as po:
        stdout = po.stdout; assert stdout is not None
        for line in stdout:
            j = json.loads(line.decode('utf8'))
            # apparently, successful runs aren't getting logged? not sure why
            jt = j.get('JOB_TYPE')
            ur = j.get('UNIT_RESULT')
            # not sure about this..
            yield j


def _unit_success_rate(unit: Unit) -> float:
    started = 0
    failed  = 0
    # TODO not sure how much time it takes to query all journals?
    for j in _unit_logs(unit):
        jt = j.get('JOB_TYPE')
        ur = j.get('UNIT_RESULT')
        if jt is not None:
            assert ur is None
            started += 1
        elif ur is not None:
            assert jt is None
            failed += 1
        else:
            # TODO eh? sometimes jobs also report Succeeded status
            # e.g. syncthing-paranoid
            pass
    if started == 0:
        assert failed == 0, unit
        return 1.0
    success = started - failed
    return success / started


def cmd_past(unit: Unit) -> None:
    mon = MonitorHelper()
    for j in _unit_logs(unit):
        ts = mon.from_usec(j['__REALTIME_TIMESTAMP'])
        msg = j['MESSAGE']
        print(ts.isoformat(), msg)


def cmd_run(*, unit: Unit, do_exec: bool) -> None:
    assert do_exec  # support without exec later
    # TODO we might have called it before via managed_units.. maybe need to cache
    states = []
    for s in systemd_state(with_body=False):
        # meh
        unit_name = s.unit_file.name
        if unit_name.endswith('.timer'):
            continue
        if s.unit_file.stem == unit:
            states.append(s)
    [state] = states
    cmdline = state.cmdline
    assert cmdline is not None
    cmds = ' '.join(map(shlex.quote, cmdline))
    logger.info(f'running: {cmds}')
    os.execvp(
        cmdline[0],
        list(cmdline),
    )


# used to use this, keeping for now just for the refernce
# def old_systemd_emailer() -> None:
#     user = getpass.getuser()
#     X = textwrap.dedent(f'''
#     [Unit]
#     Description=status email for %i to {user}
#
#     [Service]
#     Type=oneshot
#     ExecStart={SYSTEMD_EMAIL} --to {user} --unit %i --journalctl-args "-o cat"
#     # TODO why these were suggested??
#     # User=nobody
#     # Group=systemd-journal
#     ''')
#
#     write_unit(unit=f'status-email@.service', body=X, prefix=SYSTEMD_USER_DIR)
#     # I guess makes sense to reload here; fairly atomic step
#     _daemon_reload()
